#!/usr/bin/perl
#
# ripit.pl - Rips CD Audio and creates MP3 files
#            Does the following:
#		Query CDDB data for Album/Artist/Track info
#		Rip the Audio wav files from the CD
#		Encode the MP3 files
#		id3 tag the MP3 files
#		Create an M3U file
#
# Options:
#	[start_track] 		- rips from this track to last track
#	--halt 			- powers off the machine when finished
#			  	  if your configuration supports it
#	--bitrate [rate] 	- Encode MP3 at this bitrate 
#	--year [year]		- Tag MP3 with this year (included as CDDB
#				  does not store this information)
#       --device                - CDROM device to rip from
#
# Version 1.5 01/08/99 - Simon Quinn
# Version 1.5db 09/09/99 - Donald Burr <dburr@pobox.com>
#	Modified to use cdd instead of tosha, for those of us
#	(unfortunate enough) who have ATAPI/IDE CD-ROMs.  Tested
#	on FreeBSD with the FreeBSD cdd port (/usr/ports/audio/cdd),
#	YMMV if you are using a different OS.
#
# Usage: ripit.pl [--halt] [--bitrate rate] 
#                 [--year year] [--device cddevice] [start_track]
####################################################################
#
# User configurable variables
#

# Note: devices must be specified this way (strip the leading '/dev/'
# and the trailing partition letter) because this is the way cdd wants it.
$cddev 		= "acd1";		# CD Audio device
$outputdir 	= "/home/mp3/";		# Where the MP3s should go
$bitrate	= 160;			# Bitrate for MP3s
$encoder	= 0; 			# 0 - Bladeenc, 1 - Lame

$use_underscore = 0;	# Use _ instead of spaces in filenames (1 yes, 0 no)

# Track template variable. Contains the format the track names will be
# written as. The '" and "' are important and must surround the template.
# Example variables to use are: $tracknum, $trackname, $album, $artist
# E.G. Default setting produces a trackname of "02 Alphabet Street"

$tracktemplate  = '"$tracknum $trackname"';

#$tracktemplate  = '"$artist - $trackname"';

#######################################################################
# No User configurable parameters below here
######################################################################

require "flush.pl";
use Getopt::Long;

# Initialise global variables
#
$year		= "";	# Year of Audio CD - written to MP3 tag
$haltonfinish	= 0;	# Shutdown machine when finished?, default to no
$nocddbinfo	= 0;    # Assume CDDB info exists
$trackselection = "";   # Passed from command line

####################################################
# Do the following:
#  1. Get command line parameters
#  2. Get CDDB data for CD  
#  3. Get CD information, i.e. Artist, Album etc
#  4. Create Directory where the MP3s will go
#  5. Rip the CD
#  6. Encode and id3 Tag the files in the background
#  7. Create the M3U file
####################################################

# Get the parameters from the command line
if ( ! &GetOptions("halt" => \$haltonfinish, 
                  "bitrate=i" => \$bitrate,
		  "year=i" => \$year,
		  "device=s" => \$cddev) ) {
   print "Usage: ripit.pl [--halt] [--bitrate rate] [--year year]
                [--device cddevice] [start_track]\n";
   exit 1;
}

if ($haltonfinish == 1) {print "Will halt machine when finished.\n";}

# Get starting track parameter if it has been given
if ($ARGV[0] ne '') {
  $trackselection = $ARGV[0];
}              

# All logging information goes into this file
$logfile = "/tmp/ripitlog.".substr($cddev, rindex($cddev,"/") + 1);
open(RIPLOG, ">$logfile") || die "Cannot create $logfile: $!";

# All mp3 encoder messages go here
$enclog = "/tmp/encoderlog.".substr($cddev, rindex($cddev,"/") + 1);

# Calc the cdtoc to use from device, allows multiple instances of ripit
$tocfile = "/tmp/cdtoc.".substr($cddev, rindex($cddev,"/") + 1);

# Tidy up from previous session
system("rm /tmp/id3ren.log $enclog >/dev/null 2>&1");

&get_cddb();				# Get the CDDB info for the CD
&get_cdinfo();				# Extract CD info from CDDB toc
&create_seltrack($trackselection);	# Create track selection 
&create_dirs();				# Create directories MP3 files
&rip_cd();				# Rip, Encode & Tag

print "Waiting for MP3 Encoder to finish...\n";
wait; 

&create_m3u();			# Create the M3U file for the MP3 files

print "\nAll Complete.\n";
&printflush(RIPLOG,"\nAll Complete.\n");

# Delete logs and temporary files
system("rm /tmp/id3ren.log $enclog >/dev/null 2>&1");

# Shutdown and poweroff when finished only if flag is set
if ($haltonfinish == 1) {
  &printflush(RIPLOG,"Shutdown PC...\n");
  close(RIPLOG);
  system("poweroff");
}

close(RIPLOG);

exit;

#
# Create the track selection from the parameters passed
# on the command line
#
sub create_seltrack {

  my($i, $tempstr);

  ($tempstr) = @_;

  if ($tempstr =~ /-/) { die "Track selection invalid"; }

  if ($tempstr eq '') {
    for($i=1; $i <= ($#tracklist + 1); $i++) { $seltrack[$i - 1] = $i; }
  }
  elsif ($tempstr =~ /,/) { 
    @seltrack = split(/,/ , $tempstr); 
    if ($#seltrack == 0) {
      $x = 0;
      for ($i=$seltrack[0]; $i <= ($#tracklist + 1); $i++) { 
        $seltrack[$x++] = $i; 
      }
    }
  } 
  elsif ($tempstr =~ /^[0-9]*[0-9]$/) { $seltrack[0] = $tempstr; }
  else { die "Track selection invalid"; }

  # Sort the tracks in order, perl is so cool :-)
  @seltrack = sort {$a <=> $b} @seltrack;

  # Check the validaty of the track selection
  foreach (@seltrack) {
    if ($_ > ($#tracklist + 1)) {
      die "Track selection higher than number of tracks";
    }
  }

}

#
# Retrieve CDDB information using xmcd's CDA tool
# 
sub get_cddb {

  print "Getting CDDB info...\n";

  system("rm $tocfile >/dev/null 2>&1");

  system("cda -dev /dev/r${cddev}c on >/dev/null 2>&1") ;
  system("cda -dev /dev/r${cddev}c toc >$tocfile") ;
  system("cda -dev /dev/r${cddev}c off >/dev/null 2>&1") ;

  if ( ! -r $tocfile ) { die "$tocfile is missing"; }
}

#
# Read the Album, Artist, DiscID and Track Titles
# from the CDDB generated TOC file
#
sub get_cdinfo {

  local $recnum = 0;
  
  open(CDTOC, $tocfile) || die "Can't open $tocfile: $!"; 
  
  while (<CDTOC>) {
  
    chop;
   
    $recnum++ ;
  
    # Get genre and DiscID
    if ($recnum eq 1) {
      if (/Disc ID:/) {
        ($tmp1, $tmp2, $genre, $discid, $otherdiscid) = split ;
      }
      else {
        die "TOC ERROR: No Disc ID found";
      }
    }
  
    # Get the artist and album 
    if ($recnum eq 2) {
      $_ =~ s/[;:*#?|><"]//g; 			# Strip dodgey chars
      $_ =~ s/`/'/g;
      if (/\//) {
        $artist = substr($_, 0, rindex($_, "/"));
        ($artist, $album) = split (/ \/ /);
        $artist =~ s/\//-/g;
        $album =~ s/\//-/g;
        $artist = substr($artist, 0, 64);
        $album = substr($album, 0, 64);
      }
      elsif ($_ eq "(unknown disc title)" && $genre eq "(no") {
        $nocddbinfo = 1;
        $artist = "Unknown";
        $album = $otherdiscid;
      }
      elsif (length > 0) {
        # Here the artist and album are the same, i.e. no slash
        $artist = $_;
        $album = $_;
      }
      else { die "TOC ERROR: CDTOC incorrect format"; }
    }
  
    # Put track list into array
    # If there is no CDDB info then create the list 
    # with Track 01, Track 02 etc. Otherwise use the CDDB track info

    if (substr($_,1,2) >=1 && substr($_,1,2) <= 99 && substr($_,3,1) eq ' ') {
      $_ =~ s/[;:*#?|><"]//g;	# Strip dodgey chars, NB different to above
      $_ =~ s/`/'/g;		# Change ` to ' to avoid problems later
      $_ =~ s/\//-/g;		# Change / to - to avoid problems later
      if ($nocddbinfo == 0) {
        push @tracklist, substr($_, (index($_, "  ") + 2),64);
      }
      else {
        push @tracklist, "Track ".sprintf("%02d",($#tracklist + 2));
      }
    }
  
  }

  close CDTOC;

  if (($seltrack[0] - 1) > $#tracklist) {
    die "Starting track is higher than number of tracks on CD!";
  }

  if ($artist eq "") {
    &printflush(RIPLOG,"TOC ERROR: No Artist Found\n");
    die "TOC ERROR: No Artist Found";
  }

  # Display the TOC to screen
  system("cat <$tocfile");

}

#
# Create the directory where the mp3 files will go
# Fail if directory exists or can't be created
#
sub create_dirs {
  # Directory created will be: /outputdir/Artist - Album/
  # The value must end in /

  $mp3dir = $outputdir.$artist." - ".$album."/";

  if ($use_underscore == 1) { $mp3dir =~ s/ /_/g; }

  if (!opendir(TESTDIR, $mp3dir)) {
    mkdir($mp3dir,0777) || die "Cannot create directory $mp3dir: $!";
  }
  else {
    closedir(TESTDIR);
  }
}

#
# Create the track file name from the template variable
#
sub get_trackname {
    my($trnum, $trname, $riptrname);

    ($trnum, $trname) = @_;

    # Create the full file name from the track template, unless
    # the disk is unknown
    if ($nocddbinfo == 0) {
      $riptrname = $mp3dir.sprintf("%02d",$trnum).$trname;
      $tracknum = sprintf("%02d",$trnum);
      $trackname = $trname;
      if (! eval ("\$riptrname = \$mp3dir.$tracktemplate")) {
        die "Track Template is incorrect, caused eval to fail: $!" ;
      }       
    }
    else {
      $riptrname = $mp3dir.$trname;
    }

    # If using underscores then change spaces to _
    if ($use_underscore == 1) { $riptrname =~ s/ /_/g; }

    return $riptrname;

}

#
# Rip the CD 
#
sub rip_cd {

  my($startenc,$i);

  print "Ripping...\n";
  foreach (@seltrack) {

    $riptrackname = &get_trackname($_, $tracklist[$_ - 1]);
    $riptrackno = $_;

    print "Ripping track $tracklist[$_ - 1]...\n";
    &printflush(RIPLOG,"Ripping $tracklist[$_ - 1]...\n");

    if (system("cdd -t $riptrackno -q -f $cddev - 2>/dev/null | sox -t cdr - \"$riptrackname.wav\"")) {
       &printflush(RIPLOG,"cdd failed on $tracklist[$_ - 1]\n");
       die "cdd failed on $tracklist[$_ - 1]";
    }

    &printflush(RIPLOG,"Rip complete $tracklist[$_ - 1]\n");

    # Start the Encoder in the background. but only once
    if ($startenc == 0) { 
      $startenc = 1;
      unless (fork) { &enc_cd(); }
    }

  }
}

#
# Encode the wav
# This runs as a separate process from the main program which 
# allows it to continuously encode as the ripping is being done
# The encoder will also wait for the ripped wav in-case the encoder
# is faster than the CDROM, I wish :-)
#
sub enc_cd {
  my($i,$x,$ncount,$enc);


  foreach (@seltrack) {

    $riptrackname = &get_trackname($_, $tracklist[$_ - 1]);
    $ncount++;
 
    print "\nMP3 Encoding track ".$ncount." of ".($#seltrack + 1)."\n";
    &printflush(RIPLOG,"Encoding $tracklist[$_ - 1]...\n");

    # Keep looping until the file appears, ie wait for cdd
    # timeout after 30 minutes
    $x=0;
    while( ! -r "$riptrackname.wav" ){ 
      $x++; 
      if ($x > 179) { die "MP3 encoder waited 30 minutes before giving up"; } 
      sleep 10;
    }

    # Set the encoder we are going to use
    if ($encoder == 0) {
      $enc = "bladeenc -QUIT -$bitrate \"$riptrackname.wav\" ";
    }
    else {
      $enc = "lame -b $bitrate \"$riptrackname.wav\" \"$riptrackname.mp3\"";
    }

    if ( ! system("$enc >$enclog 2>&1 </dev/null")) {
 
      system("rm \"$riptrackname.wav\" >/dev/null 2>&1");

      &printflush(RIPLOG,"Encoding complete $tracklist[$_ - 1]\n");

      # Do not id3 tag if CD is unknown
      if ($nocddbinfo == 0) {
        write_id3("$riptrackname.mp3", $artist, $album, $tracklist[$_ - 1], $genre, $year);
      }
    }
    else {
      &printflush(RIPLOG,"MP3 Encoder Failed on $tracklist[$_ - 1]\n");
      die "MP3 Encoder Failed on $tracklist[$_ - 1]";
    }

  }
  exit ;
}

#
# Creates the M3U file used by players such as X11Amp
#
sub create_m3u {
  my ($file);

  $file="$artist.m3u";
  if ($use_underscore == 1) { $file =~ s/ /_/g; }

  system("cd \"$mp3dir\" ; ls -rt *.mp3 >\"$file\"");
}

# Writes the MP3 tag to the file. The standard is to use the 
# last 128 bytes in the MP3 file to store the id3tag.
# $genre must be the CDDB value
sub write_id3 {
    my($file, $artist, $album, $trackname, $genre, $year, $i);

    ($file, $artist, $album, $trackname, $genre, $year) = @_;

    # Work out the genre value. This is stored as an ascii value in the tag
    @genre_num = (0, 32, 2, 12, 80, 8, 12, 10, 16, 17, 24);
    @genre_name = ('blues', 'classical', 'country', 'data', 'folk', 'jazz', 'misc', 'newage', 'reggae', 'rock', 'soundtrack');
    for ($i=0; $i<=$#genre_name; $i++) {
      if ($genre eq $genre_name[$i]) {
        $genre = $genre_num[$i];
        last;
      }
    }

    # Write the tag information to the mp3 file
    open(IN, "+< $file") || die "MP3 tagger failed to open $file: $!";
    binmode IN	;
    sysseek IN, -128, 2	; syswrite IN, "TAG", 3;
    sysseek IN, -125, 2	; syswrite IN, padstring($trackname,30),30;
    sysseek IN, -95, 2	; syswrite IN, padstring($artist,30),30;
    sysseek IN, -65, 2	; syswrite IN, padstring($album,30),30;
    sysseek IN, -35, 2	; syswrite IN, padstring($year,4),4;
    sysseek IN, -31, 2	; syswrite IN, padstring("",30),30;
    sysseek IN, -1, 2	; syswrite IN, chr($genre), 1;
    close IN;

}

sub padstring {
  my($istr, $ilen, $i) = (shift, shift);
  if (length($istr) < $ilen) {
    for ($i=length($istr)+1; $i<=$ilen; $i++) { $istr .= "\0"; }
  }
  return $istr;
}
